package JavaML;

import java.io.*;
import java.util.Random;

/**
 * An implementation of IDataset using doubles
 * 
 * @author apocock
 */
public class DatasetDouble implements IDataset
{

  private double [][] featureMatrix;
  private int [] targetVector;
  
  private double[] dataWeights;
  
  private int numberOfFeatures;
  private int numberOfSamples;

  private int currentDataPoint;
  private int currentTarget;
  
  private boolean hasData;
  
  private int totalWeight;
  
  private int numberOfFolds;
  private int foldSize;
  private int[] foldStart;
  private int[] foldEnd;
  
  private Random rndNumberGen;
  private long rndSeed;
  
  
  /**
  * Constructor from file
  * @param  fileToLoad        is the path of the csv file to read in
  * @param  classIsFirst      specifies if the class label is at the start or the end of the file
  * @param  weightingIsFirst  specifies if the file has example weights (appears before the class label if so)
  */
  public DatasetDouble(String fileToLoad, boolean classIsFirst, boolean weightingIsFirst)
  {
    try
    {
      BufferedReader reader = new BufferedReader(new FileReader(fileToLoad));
      int newNumFeatures = 0;
      int newNumSamples = 0;
      int sampleCounter = 0;
      int countStart = 0;
      int lineLength = 0;
      double weightCounter = 0;
      
      String currentLine = reader.readLine();
      String [] lineArray = currentLine.split(",");
      if (weightingIsFirst)
      {
        newNumFeatures = lineArray.length - 2;
      }
      else
      {
        newNumFeatures = lineArray.length - 1;
      }
      newNumSamples++;
      
      while (reader.ready())
      {
        currentLine = reader.readLine();
        newNumSamples++;
      }
      
      reader.close();
      
      System.out.println("Features = " + newNumFeatures + ", Samples = " + newNumSamples);
      
      reader = new BufferedReader(new FileReader(fileToLoad));    
      featureMatrix = new double[newNumFeatures][newNumSamples];
      targetVector = new int[newNumSamples];
      dataWeights = new double[newNumSamples];
      
      numberOfFeatures = newNumFeatures;
      numberOfSamples = newNumSamples;
      
      currentLine = reader.readLine();
      lineArray = currentLine.split(",");
           
      //System.out.println("lineArray length = " + lineArray.length);
      if (weightingIsFirst)
      {
        countStart = 1;
        dataWeights[sampleCounter] = Double.parseDouble(lineArray[0]);
        weightCounter = weightCounter + dataWeights[sampleCounter];
        
      }
      
      
      if (classIsFirst)
      {
        countStart = countStart + 1;
        lineLength = lineArray.length;
        targetVector[sampleCounter] = (int) Math.floor(Double.parseDouble(lineArray[countStart-1]));
      }
      else
      {
        lineLength = lineArray.length - 1;
        targetVector[sampleCounter] = (int) Math.floor(Double.parseDouble(lineArray[lineLength]));
      }
      for (int i = countStart; i < lineLength; i++)
      {
        featureMatrix[i-countStart][sampleCounter] = Double.parseDouble(lineArray[i]);
      }
      sampleCounter++;
        
        
      while (reader.ready())
      {
        currentLine = reader.readLine();
        lineArray = currentLine.split(",");
        
        if (weightingIsFirst)
        {
          dataWeights[sampleCounter] = Double.parseDouble(lineArray[0]);
          weightCounter = weightCounter + dataWeights[sampleCounter];
        }
        
        if (classIsFirst)
        {
          lineLength = lineArray.length;
          targetVector[sampleCounter] = (int) Math.floor(Double.parseDouble(lineArray[countStart-1]));
        }
        else
        {
          lineLength = lineArray.length - 1;
          targetVector[sampleCounter] = (int) Math.floor(Double.parseDouble(lineArray[lineLength]));
        }
        for (int i = countStart; i < lineLength; i++)
        {
          featureMatrix[i-countStart][sampleCounter] = Double.parseDouble(lineArray[i]);
        }
        
        sampleCounter++;
      }
      hasData = true;
      
    
      if (!weightingIsFirst)
      {
        totalWeight = numberOfSamples;
        double weight = 1.0 / numberOfSamples;
        for (int i = 0; i < numberOfSamples; i++)
        {
          dataWeights[i] = weight;
        }
      }
      else
      {
        //normalise
        totalWeight = (int) weightCounter;
        for (int i = 0; i < numberOfSamples; i++)
        {
          dataWeights[i] = dataWeights[i] / weightCounter;
        }
      }
      numberOfFolds = 0;
      foldSize = 0;
      foldStart = new int[0];
      foldEnd = new int[0];
    }
    catch (IOException e)
    {
      System.out.println("The file to load is broken");
      hasData = false;
      featureMatrix = new double[0][0];
      targetVector = new int[0];
      numberOfFeatures = 0;
      numberOfSamples = 0;
    }
    catch (NumberFormatException e)
    {
      System.out.println("Characters where I think there should be numbers, naughty naughty");
      hasData = false;
      featureMatrix = new double[0][0];
      targetVector = new int[0];
      numberOfFeatures = 0;
      numberOfSamples = 0;
    }

    currentDataPoint = 0;
    currentTarget = 0;
    rndSeed = System.currentTimeMillis();
    rndNumberGen = new Random(rndSeed);
  }//Constructor(File)
  
  
  /**
  * Constructor from data
  * @param  newFeatureMatrix  is the new feature matrix, must have the same number of examples as <code>newTargetVector</code>
  * @param  newTargetVector   is the new list of targets
  */
  public DatasetDouble(double [][] newFeatureMatrix, int [] newTargetVector)
  {
    featureMatrix = newFeatureMatrix.clone();
    targetVector = newTargetVector.clone();
    
    numberOfFeatures = featureMatrix.length;
    numberOfSamples = targetVector.length;

    currentDataPoint = 0;
    currentTarget = 0;
    //System.out.println("Number of Features = " + featureMatrix.length + ", Number of Samples = " + targetVector.length);
    
    hasData = true;
    
    numberOfFolds = 0;
    foldSize = 0;
    foldStart = new int[0];
    foldEnd = new int[0];
    
    double weight = 1.0 / numberOfSamples;
    
    totalWeight = 1;//numberOfSamples;
    
    dataWeights = new double[numberOfSamples];
    for (int i = 0; i < numberOfSamples; i++)
    {
      dataWeights[i] = weight;
    }
  }//Constructor(double [][],double [])
  
  /**
  * Constructor from weighted data
  * @param  newFeatureMatrix  is the new feature matrix, must have the same number of examples as <code>newTargetVector</code>
  * @param  newTargetVector   is the new list of targets
  * @param  newDataWeights    is the new list of weights
  * @param  newTotalWeight    is the sum of all the weights
  */
  public DatasetDouble(double [][] newFeatureMatrix, int [] newTargetVector, double [] newDataWeights, int newTotalWeight)
  {
    featureMatrix = newFeatureMatrix.clone();
    targetVector = newTargetVector.clone();
    dataWeights = newDataWeights.clone();
    
    numberOfFeatures = featureMatrix.length;
    numberOfSamples = targetVector.length;

    currentDataPoint = 0;
    currentTarget = 0;
    
    totalWeight = newTotalWeight;
    
    hasData = true;
    
    //System.out.println("Features = " + featureMatrix.length + ", Samples = " + targetVector.length + ", Weight = " + totalWeight);
    
    numberOfFolds = 0;
    foldSize = 0;
    foldStart = new int[0];
    foldEnd = new int[0];
    
  }//Constructor(double [][],int [],double [], int)
  
  public double[][] getDataMatrixDouble()
  {
    return featureMatrix;
  }
  
  public int[] getTargetVector()
  {
    return targetVector;
  }
  
  public double[] getSampleDouble(int sampleNumber, boolean hasTarget)
  {
    double[] sample;
    if (hasTarget)
    {
      sample = new double[numberOfFeatures + 1];
    }
    else
    {
      sample = new double[numberOfFeatures];
    }
    
    for (int i = 0; i < numberOfFeatures; i++)
    {
      sample[i] = featureMatrix[i][sampleNumber];
    }
    
    if (hasTarget)
    {
      sample[numberOfFeatures] = targetVector[sampleNumber];
    }
    
    return sample;
  }
  
  public double[] getDataWeights()
  {
    return dataWeights.clone();
  }
  
  /**
  * Assigns the newWeights to the data weights.
  * Does not copy the newWeights.
  */
  public void setDataWeights(double[] newWeights)
  {
    dataWeights = newWeights;
  }
  
  public int getTotalWeight()
  {
    return totalWeight;
  }
  
  /**
  * Returns a new IDataset which contains a subset of the features of the original
  * @param  featuresRequired  is a vector of required features, all values must be < <code>numberOfFeatures</code>
  */
  public IDataset getDataFeatureSubset(int[] featuresRequired)
  {
    //returns a new IDataset which contains a subset of the features of the original
    int newLength = featuresRequired.length;
    double [][] newFeatures = new double[newLength][];
    
    for (int i = 0; i < newLength; i++)
    {
      newFeatures[i] = featureMatrix[featuresRequired[i]].clone();
    }//for the number of new features
    
    return new DatasetDouble(newFeatures, targetVector, dataWeights, totalWeight);
  }//getDataSubset(int [])
  
  /**
  * Returns a new IDataset which contains a subset of the samples of the original
  * @param  samplesRequired  is a vector of required samples, all values must be < <code>numberOfSamples</code>
  */
  public IDataset getDataSampleSubset(int[] samplesRequired)
  {
    //returns a new IDataset which contains a subset of the features of the original
    int newLength = samplesRequired.length;
    double [][] newFeatures = new double[numberOfFeatures][newLength];
    int[] newTargets = new int[newLength];
    double[] newWeights = new double[newLength];
    double newTotal = 0;
    
    for (int i = 0; i < newLength; i++)
    {
      for (int j = 0; j < numberOfFeatures; j++)
      {
        newFeatures[j][i] = featureMatrix[j][samplesRequired[i]];
      }
      newTargets[i] = targetVector[samplesRequired[i]];
      
      newWeights[i] = (int) Math.ceil(dataWeights[samplesRequired[i]] * totalWeight);
      newTotal += newWeights[i];
    }//for the number of new features
    //System.out.println("newtotal" + newTotal);
    //System.out.println("totalWeight" + totalWeight);
    
    for (int i = 0; i < newLength; i++)
    {
      newWeights[i] = newWeights[i] / newTotal;
    }
    
    return new DatasetDouble(newFeatures, newTargets, newWeights, (int)Math.round(newTotal));
  }//getDataSubset(int [])
  
  /**
  * Returns a new IDataset a sampled version of the original, based upon the supplied distribution over examples
  * @param  distributionOverFeatures  is a distribution vector, with length equal to the numberOfSamples
  * @param  sampleSize                is the number of datapoints in the new IDataset
  */
  public IDataset getSampledDataset(double[] distributionOverFeatures, int sampleSize)
  {
    double[][] sampledData = new double[numberOfFeatures][sampleSize];
    //double[] sampledWeight = new double[sampleSize];
    int[] sampledTargets = new int[sampleSize];
    
    double[] cumulativeDistribution = new double[distributionOverFeatures.length];
    
    cumulativeDistribution[0] = distributionOverFeatures[0];
    
    for (int i = 1; i < distributionOverFeatures.length; i++)
    {
      cumulativeDistribution[i] = cumulativeDistribution[i-1] + distributionOverFeatures[i];
    }//convert probability dist into cumulative prob dist
  
    //now sample from the newly normalised distribution
    double currentRnd = 0.0;
    double currentCount = 0.0;
    int j = 0;
    for (int i = 0; i < sampleSize; i++)
    {
      currentCount = 0.0;
      j = 0;
      currentRnd = rndNumberGen.nextDouble();
      
      do
      {
        j++;
      }while (currentRnd > cumulativeDistribution[j]);
      
      sampledTargets[i] = targetVector[j];
      //sampledWeight[i] = 1.0/sampleSize;
      for (int k = 0; k < numberOfFeatures; k++)
      {
        sampledData[k][i] = featureMatrix[k][j];
      }
    }//now iterate filling up the sampled data
    
    return new DatasetDouble(sampledData,sampledTargets);
  }//getSampledDataset

  public double[] getNextDataPoint() 
  {
    double [] temp = getSampleDouble(currentDataPoint,false);
    currentDataPoint++;
    if (currentDataPoint == numberOfSamples)
    {
      currentDataPoint = 0;
    }
    return temp;
  }

  public int getNextTarget() 
  {
    int temp = targetVector[currentTarget];
    currentTarget++;
    if (currentTarget == numberOfSamples)
    {
      currentTarget = 0;
    }
    return temp;
  }

  public int returnCurrentDataSampleNumber()
  {
    return currentDataPoint;
  }

  public int returnCurrentTargetSampleNumber()
  {
    return currentTarget;
  }

  public void resetDataset()
  {
    currentTarget = 0;
    currentDataPoint = 0;
  }

  /**
  * Prepares the dataset to be split into <code>newNumberOfFolds</code> folds, all previous splits are forgotten
  * MUST BE CALLED before using any other folds function
  * @param  newNumberOfFolds  is the number of folds required
  */
  public void splitIntoFolds(int newNumberOfFolds)
  {
    numberOfFolds = newNumberOfFolds;
    //now initialise fold start and end arrays
    foldStart = new int[numberOfFolds];
    foldEnd = new int[numberOfFolds];
    
    foldSize = numberOfSamples/numberOfFolds;
    System.out.println("foldSize = " + foldSize);
    
    for (int i = 0; i < numberOfFolds; i++)
    {
      foldStart[i] = i*foldSize;
      foldEnd[i] = (i+1)*foldSize;
    }//for number of folds
    
    foldEnd[numberOfFolds-1] = numberOfSamples;
    
    System.out.println("last fold is " + (foldEnd[numberOfFolds-1] - foldStart[numberOfFolds-1]));
  }//splitIntoFolds(int)
  
  private double[][] getFoldData(int foldNumber)
  {
    if (foldNumber < numberOfFolds)
    {
      int start = foldStart[foldNumber];
      int end = foldEnd[foldNumber];
      
      int foldLength = end - start;
      double[][] foldData = new double[numberOfFeatures][foldLength];
      
      for (int i = 0; i < numberOfFeatures; i++)
      {
        for (int j = start; j < end; j++)
        {
          foldData[i][j-start] = featureMatrix[i][j];
        }//iterate over fold samples
      }//iterate over features
      
      return foldData;
    }
    else
    {
      throw new RuntimeException("foldNumber >= numberOfFolds");
    }
  }//getFoldData(int)
  
  private int[] getFoldTargets(int foldNumber)
  {
    if (foldNumber < numberOfFolds)
    {
      int start = foldStart[foldNumber];
      int end = foldEnd[foldNumber];
      
      int foldLength = end - start;
      int[] foldTargets = new int[foldLength];
      
      for (int j = start; j < end; j++)
      {
        foldTargets[j-start] = targetVector[j];
      }//iterate over fold samples
      
      return foldTargets;
    }
    else
    {
      throw new RuntimeException("foldNumber >= numberOfFolds");
    }  
  }//getFoldTargets(int)
  
  private double[] getFoldWeight(int foldNumber)
  {
    if (foldNumber < numberOfFolds)
    {
      int start = foldStart[foldNumber];
      int end = foldEnd[foldNumber];
      
      int foldLength = end - start;
      double[] foldWeight = new double[foldLength];
      
      for (int j = start; j < end; j++)
      {
        foldWeight[j-start] = dataWeights[j];
      }//iterate over fold samples
      
      return foldWeight;
    }
    else
    {
      throw new RuntimeException("foldNumber >= numberOfFolds");
    }  
  }//getFoldWeight(int)
  
  private double[][] getOtherFoldsData(int foldNumber)
  {
    if (foldNumber < numberOfFolds)
    {
      int start = foldStart[foldNumber];
      int end = foldEnd[foldNumber];
      int midStart = start - end;
      
      int foldLength = end - start;
      double[][] foldData = new double[numberOfFeatures][numberOfSamples-foldLength];
      
      for (int i = 0; i < numberOfFeatures; i++)
      {
        for (int j = 0; j < start; j++)
        {
          foldData[i][j] = featureMatrix[i][j];
        }//iterate over fold samples
        
        for (int j = end; j < numberOfSamples; j++)
        {
          foldData[i][j+midStart] = featureMatrix[i][j];
        }//iterate over fold samples
      }//iterate over features
      
      return foldData;
    }
    else
    {
      throw new RuntimeException("foldNumber >= numberOfFolds");
    }  
  }//getOtherFoldsData(int)
  
  private int[] getOtherFoldsTargets(int foldNumber)
  {
    if (foldNumber < numberOfFolds)
    {
      int start = foldStart[foldNumber];
      int end = foldEnd[foldNumber];
      int midStart = start - end;
      
      int foldLength = end - start;
      int[] foldData = new int[numberOfSamples-foldLength];
      
      for (int j = 0; j < start; j++)
      {
        foldData[j] = targetVector[j];
      }//iterate over fold samples
      
      for (int j = end; j < numberOfSamples; j++)
      {
        foldData[j+midStart] = targetVector[j];
      }//iterate over fold samples
      
      return foldData;
    }
    else
    {
      throw new RuntimeException("foldNumber >= numberOfFolds");
    }  
  }//getOtherFoldsTargets(int)
  
  private double[] getOtherFoldsWeight(int foldNumber)
  {
    if (foldNumber < numberOfFolds)
    {
      int start = foldStart[foldNumber];
      int end = foldEnd[foldNumber];
      int midStart = start - end;
      
      int foldLength = end - start;
      double[] foldData = new double[numberOfSamples-foldLength];
      
      for (int j = 0; j < start; j++)
      {
        foldData[j] = dataWeights[j];
      }//iterate over fold samples
      
      for (int j = end; j < numberOfSamples; j++)
      {
        foldData[j+midStart] = dataWeights[j];
      }//iterate over fold samples
      
      return foldData;
    }
    else
    {
      throw new RuntimeException("foldNumber >= numberOfFolds");
    }  
  }//getOtherFoldsWeight(int)
  

  /**
  * Returns a new dataset without fold <code>foldNumber</code>
  * @param  foldNumber  is the current testing fold
  */
  public IDataset returnWeightedTrainingSet(int foldNumber)
  {
    double[][] otherData = this.getOtherFoldsData(foldNumber);
    double[] otherWeight = this.getOtherFoldsWeight(foldNumber);
    int[] otherTargets = this.getOtherFoldsTargets(foldNumber);
    
    //now calculate the fold weight
    
    int otherTotalWeight = 0;
    int temp = 0;
    double[] normalisedWeights = new double[otherWeight.length];
    
    for (int i = 0; i < otherWeight.length; i++)
    {
      temp = (int) Math.round(otherWeight[i] * totalWeight);
      otherTotalWeight = otherTotalWeight + temp;
      normalisedWeights[i] = temp;
    }
    
    for (int i = 0; i < otherWeight.length; i++)
    {
      normalisedWeights[i] = normalisedWeights[i] / otherTotalWeight;
    }
    
    //renormalise weights
    
    //System.out.println("Training Weight = " + otherTotalWeight);
    
    return new DatasetDouble(otherData,otherTargets,normalisedWeights,otherTotalWeight);
  }//returnWeightedTrainingSet(int)
  

  /**
  * Returns a new dataset containing fold <code>foldNumber</code>
  * @param  foldNumber  is the current testing fold
  */
  public IDataset returnWeightedTestingSet(int foldNumber)
  {
    double[][] data = this.getFoldData(foldNumber);
    double[] weight = this.getFoldWeight(foldNumber);
    int[] targets = this.getFoldTargets(foldNumber);
    
    //now calculate the fold weight
    
    int newTotalWeight = 0;
    int temp = 0;
    double[] normalisedWeights = new double[weight.length];
    
    for (int i = 0; i < weight.length; i++)
    {
      temp = (int) Math.round(weight[i] * totalWeight);
      newTotalWeight = newTotalWeight + temp;
      normalisedWeights[i] = temp;
    }
    
    for (int i = 0; i < weight.length; i++)
    {
      normalisedWeights[i] = normalisedWeights[i] / newTotalWeight;
    }
    
    //renormalise weights
    
    //System.out.println("Testing Weight = " + newTotalWeight);
    //System.out.println("Total Weight = " + totalWeight);
    
    return new DatasetDouble(data,targets,normalisedWeights,newTotalWeight);
  }//returnWeightedTestingSet(int)
  
  /**
  * Returns a new dataset without fold <code>foldNumber</code>, by sampling from the weight distribution
  * @param  foldNumber  is the current testing fold
  * @param  sampleSize  is the number of examples in the new dataset
  */
  public IDataset returnSampledTrainingSet(int foldNumber, int sampleSize)
  {
    //first extract the correct folds
    //then normalise the weights
    //then sample from the distribution a dataset of size sampleSize
    
    double[][] otherData = this.getOtherFoldsData(foldNumber);
    double[] otherWeight = this.getOtherFoldsWeight(foldNumber);
    int[] otherTargets = this.getOtherFoldsTargets(foldNumber);
    
    double[][] sampledData = new double[numberOfFeatures][sampleSize];
    //double[] sampledWeight = new double[sampleSize];
    int[] sampledTargets = new int[sampleSize];
    
    //now calculate the fold weight
    
    int otherTotalWeight = 0;
    int temp = 0;
    double[] normalisedWeights = new double[otherWeight.length];
    
    for (int i = 0; i < otherWeight.length; i++)
    {
      temp = (int) Math.round(otherWeight[i] * totalWeight);
      otherTotalWeight = otherTotalWeight + temp;
      normalisedWeights[i] = temp;
    }
    
    //renormalise weights
    for (int i = 0; i < otherWeight.length; i++)
    {
      normalisedWeights[i] = normalisedWeights[i] / otherTotalWeight;
    }
    
    //now sample from the newly normalised distribution
    double currentRnd = 0.0;
    double currentCount = 0.0;
    int j = 0;
    for (int i = 0; i < sampleSize; i++)
    {
      currentCount = 0.0;
      j = 0;
      currentRnd = rndNumberGen.nextDouble();
      
      while ((j < otherWeight.length) && (currentCount < currentRnd))
      {
        currentCount += otherWeight[j];
        j++;
      }
      j--;
      
      sampledTargets[i] = otherTargets[j];
      //sampledWeight[i] = 1.0/sampleSize;
      for (int k = 0; k < numberOfFeatures; k++)
      {
        sampledData[k][i] = otherData[k][j];
      }
    }//now iterate filling up the sampled data
    
    return new DatasetDouble(sampledData,sampledTargets);
  }//returnSampledTrainingSet(int)
  
  
  /**
  * Returns a new dataset containing fold <code>foldNumber</code>, by sampling from the weight distribution
  * @param  foldNumber  is the current testing fold
  * @param  sampleSize  is the number of examples in the new dataset
  */
  public IDataset returnSampledTestingSet(int foldNumber, int sampleSize)
  {
    //first extract the correct fold
    //then normalise the weights
    //then sample from the distribution a dataset of size sampleSize
    
    double[][] data = this.getFoldData(foldNumber);
    double[] weight = this.getFoldWeight(foldNumber);
    int[] targets = this.getFoldTargets(foldNumber);
    
    double[][] sampledData = new double[numberOfFeatures][sampleSize];
    //double[] sampledWeight = new double[sampleSize];
    int[] sampledTargets = new int[sampleSize];
    
    //now calculate the fold weight
    
    int newTotalWeight = 0;
    int temp = 0;
    double[] normalisedWeights = new double[weight.length];
    
    for (int i = 0; i < weight.length; i++)
    {
      temp = (int) Math.round(weight[i] * totalWeight);
      newTotalWeight = newTotalWeight + temp;
      normalisedWeights[i] = temp;
    }
    
    for (int i = 0; i < weight.length; i++)
    {
      normalisedWeights[i] = normalisedWeights[i] / newTotalWeight;
    }
    
    //now sample from the newly normalised distribution
    double currentRnd = 0.0;
    double currentCount = 0.0;
    int j = 0;
    for (int i = 0; i < sampleSize; i++)
    {
      currentCount = 0.0;
      j = 0;
      currentRnd = rndNumberGen.nextDouble();
      
      while ((j < weight.length) && (currentCount < currentRnd))
      {
        currentCount += weight[j];
        j++;
      }
      
      j--;
      
      sampledTargets[i] = targets[j];
      //sampledWeight[i] = 1.0/sampleSize;
      for (int k = 0; k < numberOfFeatures; k++)
      {
        sampledData[k][i] = data[k][j];
      }
    }//now iterate filling up the sampled data
    
    return new DatasetDouble(sampledData,sampledTargets);
  }//returnSampledTestingSet(int)
  
   
  public void setRandomSeed(long newSeed)
  {
    rndNumberGen.setSeed(newSeed);
    rndSeed = newSeed;
  }  
  
  public DataType getDataType()
  {
    return DataType.Double;
  }
  
  public int getNumberOfFeatures()
  {
    return numberOfFeatures;
  }
  
  public int getNumberOfSamples()
  {
    return numberOfSamples;
  }
  
  public boolean hasData()
  {
    return hasData;
  }

}//class DatasetDouble
