package JavaML;

import java.util.HashMap;

/**
 * An Online implementation of a two-class Naive Bayes Classifier, 
 *
 * Compiler warnings are due to the generics implementation in Java not allowing arrays of a generic collection
 *  
 * @author apocock
 */
public class NaiveBayes implements IOnlineClassifier
{  
  private double positivePrior;
  private double negativePrior;
  
  private static int initialHashSize = 10;
  
  /* because of the terrible implementation of generics, i can't have an array of HashMap<K,V>
  ** got to use the unchecked type, which generates lots of warnings
  */
  
  //hash map is <K=double,V=integer>
  private HashMap[] pKeyGivenTruth;
  private HashMap[] pKeyGivenFalse;
  private int totalTruth;
  private int totalFalse;
  
  private boolean pseudoCounts;
  
  private int numberOfFeatures;
  
  private boolean isTrained;
  
  /**
  * Constructor, by default without psuedoCounts
  */
  public NaiveBayes()
  {
    positivePrior = 0.5;
    negativePrior = 0.5;
    pseudoCounts = false;
  }//default constructor
  
  /**
  * Constructor, pseudoCounts is set to the value of newCounts
  * @param  newCounts   a boolean value for the pseudoCount implementation
  */
  public NaiveBayes(boolean newCounts)
  {
    positivePrior = 0.5;
    negativePrior = 0.5;
    pseudoCounts = newCounts;
  }//default constructor

  /**
  * Copy constructor, used by the ensemble algorithms
  */
  public IOnlineClassifier copyClassifier()
  {
    IOnlineClassifier newNBayes = new NaiveBayes(pseudoCounts);
    return newNBayes;
  }//copyClassifier
  
  /**
  * setupClassifier is used to modify classifiers after they have been constructed through the copy constructor,
  * at that point you could either do reflection to figure out the class and reconstruct it, or you could pass it
  * a string of arguments and have it reconfigure itself.
  *
  * This classifier accepts "PseudoCounts=<true,false>"
  *
  * @param  arguments   a space separated list of parameter=value tuples.
  */
  public void setupClassifier(String arguments)
  {
    String [] argsList = arguments.split(" ");
    try
    {
      for (int i = 0; i < argsList.length; i++)
      {
        String [] arg = argsList[i].split("=");
        if (arg.length == 2)
        {
          if (arg[0].equals("PseudoCounts"))
          {
            pseudoCounts = Boolean.parseBoolean(arg[1]);
          }//if weaklearner argument
        }//if arg formats
      }//for number of arguments
    }//try
    catch (NumberFormatException e)
    {
      System.out.println("Incorrectly formatted number in arguments");
    }
    catch (ArrayIndexOutOfBoundsException e)
    {
      System.out.println("Incorrectly formatted argument, missing equals");
    }
    catch (Exception e)
    {
      System.out.println("Other generic exception in setupClassifier");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
  }//setupClassifier(String)

  /**
  * Is the main offline train function, which takes an dataset and trains on it
  * 
  * in this classifier, iterations does nothing
  *
  * @param  trainingData    A training dataset
  * @param  iterations      Does nothing (there for IClassifier support for Boosting)
  */
  public double train(IDataset trainingData, int iterations)
  {
    double returnVal = 0.0;
        
    int numberOfSamples = trainingData.getNumberOfSamples();
    numberOfFeatures = trainingData.getNumberOfFeatures();
    double [][] featureMatrix = trainingData.getDataMatrixDouble();
    int [] labels = trainingData.getTargetVector();
    double [] sampleWeights = trainingData.getDataWeights();
    int [] intWeights = new int[numberOfSamples];
    int totalWeight = trainingData.getTotalWeight();
    
    totalFalse = 0;
    totalTruth = 0;
    pKeyGivenTruth = new HashMap[numberOfFeatures];
    pKeyGivenFalse = new HashMap[numberOfFeatures];
    
    for(int i = 0; i < numberOfSamples; i++)
    {
      intWeights[i] = (int) Math.ceil(sampleWeights[i] * totalWeight);
      if (labels[i] == IClassifier.positiveClass)
      {
        totalTruth += intWeights[i];
      }
      else
      {
        totalFalse += intWeights[i];
      }
    }//for samples
    
    for (int i = 0; i < numberOfFeatures; i++)
    {
      pKeyGivenTruth[i] = new HashMap(initialHashSize);
      pKeyGivenFalse[i] = new HashMap(initialHashSize);
      for(int j = 0; j < numberOfSamples; j++)
      {
        if (labels[j] == IClassifier.positiveClass)
        {
          Double key = featureMatrix[i][j];
          Integer value = (Integer) pKeyGivenTruth[i].remove(key);
          if (value == null)
          {
            value = intWeights[j];
          }
          else
          {
            value = value + intWeights[j];
          }
          pKeyGivenTruth[i].put(key,value);
        }
        else if (labels[j] == IClassifier.negativeClass)
        {
          Double key = featureMatrix[i][j];
          Integer value = (Integer) pKeyGivenFalse[i].remove(key);
          if (value == null)
          {
            value = intWeights[j];
          }
          else
          {
            value = value + intWeights[j];
          }
          pKeyGivenFalse[i].put(key,value);
        }
      }//for samples
    }//for features
    
    positivePrior = (double) totalTruth / (totalTruth + totalFalse);
    negativePrior = 1.0 - positivePrior;
    
    isTrained = true;
       
    return test(trainingData);
    //return returnVal;
  }//train(IDataset, int)
  
  /**
  * Is the online train function, which takes a single example and trains on it
  * then returns the classification.
  *
  * @param  trainingData    A training sample
  * @param  target          class label for the sample
  */
  public int onlineTrain(double[] trainingData, int target)
  {    
    if (!isTrained)
    {
      numberOfFeatures = trainingData.length;
      totalFalse = 0;
      totalTruth = 0;
      pKeyGivenTruth = new HashMap[numberOfFeatures];
      pKeyGivenFalse = new HashMap[numberOfFeatures];
      for (int i = 0; i < numberOfFeatures; i++)
      {
        pKeyGivenTruth[i] = new HashMap(initialHashSize);
        pKeyGivenFalse[i] = new HashMap(initialHashSize);
      }
      isTrained = true;
    }//if untrained
    
    if (target == IClassifier.positiveClass)
    {
      totalTruth += 1;
    }
    else
    {
      totalFalse += 1;
    }
    
    for (int i = 0; i < numberOfFeatures; i++)
    {
      if (target == IClassifier.positiveClass)
      {
        Double key = trainingData[i];
        Integer value = (Integer) pKeyGivenTruth[i].remove(key);
        if (value == null)
        {
          value = 1;
        }
        else
        {
          value = value + 1;
        }
        pKeyGivenTruth[i].put(key,value);
      }
      else if (target == IClassifier.negativeClass)
      {
        Double key = trainingData[i];
        Integer value = (Integer) pKeyGivenFalse[i].remove(key);
        if (value == null)
        {
          value = 1;
        }
        else
        {
          value = value + 1;
        }
        pKeyGivenFalse[i].put(key,value);
      }
    }//for features
    
    positivePrior = (double) totalTruth / (totalTruth + totalFalse);
    negativePrior = 1.0 - positivePrior;
    
    return classifySample(trainingData);
  }

  /**
  * Is the main offline test function, which takes an dataset and returns the testing error
  * on that dataset
  * 
  * @param  testingData    A testing dataset
  */
  public double test(IDataset testingData)
  {
    double returnVal = 0.0;
    if (isTrained)
    {
      //run test
      int numberOfTestSamples = testingData.getNumberOfSamples();
      int[] targetVector = testingData.getTargetVector();
      double[] dataWeights = testingData.getDataWeights();
      
      for (int i = 0; i < numberOfTestSamples; i++)
      {
        double[] currentSample = testingData.getSampleDouble(i,false);
        int finalClassification = classifySample(currentSample);
        if (finalClassification != targetVector[i])
        {
          returnVal += dataWeights[i];
        }
      }//for number of testing samples
      
      //returnVal = returnVal / numberOfTestSamples;
    }//if trained
    else
    {
      returnVal = -1.0;
    }//else untrained, return -1
    return returnVal;
  }//test(IDataset)
  
  /**
  * Returns the classifications made by the trained classifier on the testing data
  * 
  * @param  testingData    A testing dataset
  */
  public int[] classifyDataset(IDataset testingData)
  {
    int[] classifications;
  
    if (isTrained)
    {
      int dataNumberOfFeatures = testingData.getNumberOfFeatures();
      int numberOfSamples = testingData.getNumberOfSamples();
      classifications = new int[numberOfSamples];
      double[][] featureMatrixPtr = testingData.getDataMatrixDouble();
      
      if (dataNumberOfFeatures != numberOfFeatures)
      {
        System.out.println("dataFeatures = " + dataNumberOfFeatures + ", features = " + numberOfFeatures);
      }
      
      for (int i = 0; i < numberOfSamples; i++)
      {
        double posClass = 1.0;
        double negClass = 1.0;
        
        for (int j = 0; j < numberOfFeatures; j++)
        {
          Integer tmp = (Integer) pKeyGivenTruth[j].get(featureMatrixPtr[j][i]);
          if (tmp != null)
          {
            posClass *= (double) tmp.intValue() / totalTruth;
          }
          else
          {
            if (pseudoCounts)
            {
              posClass *= 0.5 / totalTruth;
            }
            else
            {
              posClass = 0.0;
            }
          }
          
          tmp = (Integer) pKeyGivenFalse[j].get(featureMatrixPtr[j][i]);
          if (tmp != null)
          {
            negClass *= (double) tmp.intValue() / totalFalse;
          }
          else
          {
            if (pseudoCounts)
            {
              negClass *= 0.5 / totalFalse;
            }
            else
            {
              negClass = 0.0;
            }
          }
        }
        posClass *= positivePrior;
        negClass *= negativePrior;
        
        double probTruth = posClass / (posClass + negClass);
        
        
        if (probTruth > 0.5)
        {
          classifications[i] = IClassifier.positiveClass;
        }
        else
        {
          classifications[i] = IClassifier.negativeClass;
        }
      }//for loop over number of samples
    }
    else
    {
      classifications = null;
    }
      
    return classifications;
  }//classifyDataset(IDataset)
        
  
  /**
  * Returns the classification for a single example
  * 
  * @param  sample        A testing sample
  */  
  public int classifySample(double[] sample)
  {
    int classification;
  
    if (isTrained)
    {        
      double posClass = 1.0;
      double negClass = 1.0;
      
      for (int i = 0; i < numberOfFeatures; i++)
      {
        Integer tmp = (Integer) pKeyGivenTruth[i].get(sample[i]);
        if (tmp != null)
        {
          posClass *= (double) tmp.intValue() / totalTruth;
        }
        else
        {
          if (pseudoCounts)
          {
            posClass *= 0.5 / totalTruth;
          }
          else
          {
            posClass = 0.0;
          }
        }
        
        tmp = (Integer) pKeyGivenFalse[i].get(sample[i]);
        if (tmp != null)
        {
          negClass *= (double) tmp.intValue() / totalFalse;
        }
        else
        {
          if (pseudoCounts)
          {
            negClass *= 0.5 / totalFalse;
          }
          else
          {
            negClass = 0.0;
          }
        }
      }
      posClass *= positivePrior;
      negClass *= negativePrior;
      
      double probTruth = posClass / (posClass + negClass);
      
      if (probTruth > 0.5)
      {
        classification = IClassifier.positiveClass;
      }
      else
      {
        classification = IClassifier.negativeClass;
      }
    }
    else
    {
      classification = 0;
    }
      
    return classification;
  }//classifySample(double[])
  
  public int[] getFeatureList()
  {
    if (isTrained)
    {
      int [] returnArray = new int[0];
      
      //returnArray[0] = featureToUse;
      
      return returnArray;
    }//if trained
    else
    {
      return new int[0];
    }//else untrained, return empty list
  }//getFeatureList()
  
  public boolean getIsTrained()
  {
    return isTrained;
  }
  
  public boolean supportsWeightedData()
  {
    return true;
  }
  
  public boolean supportsMultiClassData()
  {
    return false;
  }
}//public class ThresholdClassifier
