package JavaML;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import cern.jet.random.Poisson;

/**
 * An implementation of ONSBoost. (http://dx.doi.org/10.1007/978-3-642-12127-2_21)
 * When using offline datasets, behaves like FloatBoost.
 * Executing the train() function resets the classifier before training.
 *  
 * @author apocock
 */
public class ONSBoost implements IOnlineClassifier
{
  private int numberOfWeakLearners;
  
  private double[] errorValues;
  private IOnlineClassifier[] weakLearners;
  private double[] weakLearnerCorrect;
  private double[] weakLearnerWrong;
  
  private double[] errors;
  
  private IOnlineClassifier rootWeakLearner;
  private String weakLearnerArgs;
  private int weaklearnerIterations;
  
  //parameters
  private int examplesBeforeBackwardStep;
  private boolean usePretrainedClassifiers;
  private boolean storeOldClassifiers;
  private boolean protectNewClassifiers;
  private int examplesToProtect;
  private boolean floatNumberOfClassifiers;
  private int maxNumClassifiers;
  private int minNumClassifiers;
  private boolean decayOldClassifiers;
  private int examplesToDecay;
  private int windowSize;
  private boolean isWindowed;
  private boolean isWindowAveraging;
  private boolean isIntervalSampling;
  
  //internal values for parameters
  private int onlineExamplesCounter;
  private int[] examplesOfProtection;
  private int windowCounter;
  private int[][] windowLearners;
  private double[][] windowAlphas;
  private double[] windowOutput;
  private int[] windowTarget;
  
  
  private boolean isPrimed;
  private boolean isTrained;
  
  /**
  * Constructor, needs to be supplied with a weak learner before training
  *
  */
  public ONSBoost()
  {
    numberOfWeakLearners = 0;
    
    errorValues = new double[0];
    weakLearners = new IOnlineClassifier[0];
    weakLearnerCorrect = new double[0];
    weakLearnerWrong = new double[0];
    
    rootWeakLearner = null;
    weakLearnerArgs = "";
    weaklearnerIterations = -1;
    
    defaultSetup();
  }//OnlineFloatBoost constructor
  
  /**
  * Constructor, initialises everything required
  * @param  numLearners   Is the number of weak learners to use
  * @param  rootLearner   Is an IOnlineClassifier which all the weak learners are constructed from
  * @param  learnerArgs   Are any arguments that weak learner needs to be constructed with
  */
  public ONSBoost(int numLearners, IOnlineClassifier rootLearner, String learnerArgs)
  {
    numberOfWeakLearners = numLearners;
    
    errorValues = new double[numberOfWeakLearners];
    weakLearners = new IOnlineClassifier[numberOfWeakLearners];
    weakLearnerCorrect = new double[numberOfWeakLearners];
    weakLearnerWrong = new double[numberOfWeakLearners];
    
    rootWeakLearner = rootLearner;
    weakLearnerArgs = learnerArgs;
    weaklearnerIterations = -1;
    
    defaultSetup();
  }//constructor
  
  /**
  * Constructor, initialises everything required
  * @param  numLearners   Is the number of weak learners to use
  * @param  rootLearner   Is an IOnlineClassifier which all the weak learners are constructed from
  * @param  learnerArgs   Are any arguments that weak learner needs to be constructed with
  * @param  learnerIterations If the weak learner needs the iterations parameter (e.g. for boosting a neural net)
  */
  public ONSBoost(int numLearners, IOnlineClassifier rootLearner, int learnerIterations, String learnerArgs)
  {
    numberOfWeakLearners = numLearners;
    
    errorValues = new double[numberOfWeakLearners];
    weakLearners = new IOnlineClassifier[numberOfWeakLearners];
    weakLearnerCorrect = new double[numberOfWeakLearners];
    weakLearnerWrong = new double[numberOfWeakLearners];
    
    rootWeakLearner = rootLearner;
    weakLearnerArgs = learnerArgs;
    weaklearnerIterations = learnerIterations;
    
    defaultSetup();
  }//constructor
  
  private void defaultSetup()
  {
    isPrimed = false;
    isTrained = false;
    
    onlineExamplesCounter = 0;
    examplesBeforeBackwardStep = 50;
  
    usePretrainedClassifiers = false;
    storeOldClassifiers = false;
    
    protectNewClassifiers = true;
    examplesToProtect = 100;
    examplesOfProtection = new int[numberOfWeakLearners];
    
    floatNumberOfClassifiers = false;
    maxNumClassifiers = numberOfWeakLearners;
    minNumClassifiers = numberOfWeakLearners;
    
    decayOldClassifiers = false;
    examplesToDecay = 0;
    
    windowSize = 50;
    isWindowAveraging = false;
    isWindowed = true;
    isIntervalSampling = false;
    windowCounter = 0;
    
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      weakLearners[i] = rootWeakLearner.copyClassifier();
      errorValues[i] = 0.5;
      examplesOfProtection[i] = examplesToProtect;
    }
  }//defaultSetup()
  
  /**
  * Copy constructor
  */
  public IOnlineClassifier copyClassifier()
  {
    IOnlineClassifier newONSBoost = new ONSBoost(numberOfWeakLearners, rootWeakLearner, weaklearnerIterations, weakLearnerArgs);
    
    newONSBoost.setupClassifier(this.generateClassifierSetup());
    
    return newONSBoost;
  }//copyClassifier()
  
  
  /**
  * setupClassifier is used to modify classifiers after they have been constructed through the copy constructor,
  * at that point you could either do reflection to figure out the class and reconstruct it, or you could pass it
  * a string of arguments and have it reconfigure itself.
  *
  * This classifier accepts a long list of arguments
  *       - "WeakLearner=<CLASSNAME>"
  *       - "NumLearners=<int>"
  *       - "WeakLearnerIterations=<int>"
  *       - "ProtectNewClassifiers=<true,false>"
  *       - "ExamplesToProtect=<int>"
  *       - "WindowSize=<int>"
  *
  * @param  arguments   a space separated list of parameter=value tuples.
  */
  public void setupClassifier(String arguments)
  {
    String [] argsList = arguments.split(" ");
    try
    {
      for (int i = 0; i < argsList.length; i++)
      {
        String [] arg = argsList[i].split("=");
        if (arg.length == 2)
        {
          if (arg[0].equals("WeakLearner"))
          {
            Class learnerClass = null;
            
            System.out.println(arg[1]);
            
            learnerClass = Class.forName(arg[1]);
            
            Object learnerObject = learnerClass.newInstance();
            
            rootWeakLearner = (IOnlineClassifier) learnerClass.cast(learnerObject);
          }//if weaklearner argument
          else if (arg[0].equals("NumLearners"))
          {
            int numLearners = Integer.parseInt(arg[1]);
            numberOfWeakLearners = numLearners;
            errorValues = new double[numberOfWeakLearners];
            weakLearners = new IOnlineClassifier[numberOfWeakLearners];
          }
          else if (arg[0].equals("WeakLearnerIterations"))
          {
            weaklearnerIterations = Integer.parseInt(arg[1]);
          }
          else if (arg[0].equals("ExamplesBeforeBackwardStep"))
          {
            examplesBeforeBackwardStep = Integer.parseInt(arg[1]);
          }
          else if (arg[0].equals("UsePretrainedClassifiers"))
          {
            usePretrainedClassifiers = Boolean.parseBoolean(arg[1]);
          }
          else if (arg[0].equals("StoreOldClassifiers"))
          {
            storeOldClassifiers = Boolean.parseBoolean(arg[1]);
          }
          else if (arg[0].equals("ProtectNewClassifiers"))
          {
            protectNewClassifiers = Boolean.parseBoolean(arg[1]);
          }
          else if (arg[0].equals("ExamplesToProtect"))
          {
            examplesToProtect = Integer.parseInt(arg[1]);
          }
          else if (arg[0].equals("FloatNumberOfClassifiers"))
          {
            floatNumberOfClassifiers = Boolean.parseBoolean(arg[1]);
          }
          else if (arg[0].equals("MaxNumberOfClassifiers"))
          {
            maxNumClassifiers = Integer.parseInt(arg[1]);
          }
          else if (arg[0].equals("MinNumberOfClassifiers"))
          {
            minNumClassifiers = Integer.parseInt(arg[1]);
          }
          else if (arg[0].equals("DecayOldClassifiers"))
          {
            decayOldClassifiers = Boolean.parseBoolean(arg[1]);
          }
          else if (arg[0].equals("ExamplesToDecay"))
          {
            examplesToDecay = Integer.parseInt(arg[1]);
          }
          else if(arg[0].equals("WindowSize"))
          {
            windowSize = Integer.parseInt(arg[1]);
          }
          else if (arg[0].equals("isWindowAveraging"))
          {
            isWindowAveraging = Boolean.parseBoolean(arg[1]);
            if (isWindowAveraging)
            {
              isWindowed = false;
              isIntervalSampling = false;
            }
          }
          else if (arg[0].equals("isWindowed"))
          {
            isWindowed = Boolean.parseBoolean(arg[1]);
            if (isWindowed)
            {
              isWindowAveraging = false;
              isIntervalSampling = false;
            }
          }
          else if (arg[0].equals("isIntervalSampling"))
          {
            isIntervalSampling = Boolean.parseBoolean(arg[1]);
            if (isIntervalSampling)
            {
              isWindowAveraging = false;
              isWindowed = false;
            }
          }
          else
          {
            System.out.println("Unrecognised argument: " + arg[0] + " with value = " + arg[1]);
          }
        }//if arg formats
      }//for number of arguments
    }//try
    catch (NumberFormatException e)
    {
      System.out.println("Incorrectly formatted number in arguments");
    }
    catch (ArrayIndexOutOfBoundsException e)
    {
      System.out.println("Incorrectly formatted argument, missing equals");
    }
    catch (ClassCastException e)
    {
      System.out.println("Class casting exception, the WeakLearner did not implement IClassifier");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
    catch (ClassNotFoundException e)
    {
      System.out.println("WeakLearner class not found");
    }
    catch (InstantiationException e)
    {
      System.out.println("Instantiation exception on creation of new WeakLearner");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
    catch (Exception e)
    {
      System.out.println("Other generic exception in setupClassifier");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
  }//setupClassifier
  
  private String generateClassifierSetup()
  {
    String returnData = "";
    
    returnData = "ExamplesBeforeBackwardStep=" + examplesBeforeBackwardStep;
    
    returnData += "UsePretrainedClassifiers=" + usePretrainedClassifiers;
    
    returnData += "StoreOldClassifiers=" + storeOldClassifiers;
    
    returnData += "ProtectNewClassifiers=" + protectNewClassifiers;
    
    returnData += "ExamplesToProtect=" + examplesToProtect;
    
    returnData += "FloatNumberOfClassifiers=" + floatNumberOfClassifiers;
    
    returnData += "MaxNumberOfClassifiers=" + maxNumClassifiers;
    
    returnData += "MinNumberOfClassifiers=" + minNumClassifiers;
    
    returnData += "DecayOldClassifiers=" + decayOldClassifiers;
    
    returnData += "ExamplesToDecay=" + examplesToDecay;
    
    returnData += "WindowSize=" + windowSize;
    
    returnData += "isWindowAveraging=" + isWindowAveraging;
    
    returnData += "isWindowed=" + isWindowed;
    
    //returnData += "";
    
    return returnData;
  }//generateClassifierSetup()

  /**
  * Is the main offline train function, which takes an dataset and trains on it using the FloatBoost algorithm
  * 
  *
  * @param  trainingData    A training dataset
  * @param  iterations      overrides the inbuilt number of weak leaners, and creates <code>iterations</code> weak learners
  */
  public double train(IDataset trainingData, int iterations)
  {
    double returnVal = 0.0;
    if ((rootWeakLearner == null) || (iterations < 1))
    {
      returnVal = -1.0;
    }
    else
    {
      returnVal = offlineTrain(trainingData,iterations);
    }
    return returnVal;
  }
  
  
  /**
  * Is the online train function, which takes a single example and trains on it
  * then returns the classification.
  *
  * @param  trainingData    A training sample
  * @param  target          class label for the sample
  */
  public int onlineTrain(double[] trainingData, int target)
  {    
    if (numberOfWeakLearners == 0)
    {
      errorValues = new double[numberOfWeakLearners];
      weakLearners = new IOnlineClassifier[numberOfWeakLearners];
      weakLearnerCorrect = new double[numberOfWeakLearners];
      weakLearnerWrong = new double[numberOfWeakLearners];
    }
    
    if (windowLearners == null)
    {
      windowLearners = new int[windowSize][numberOfWeakLearners];
      windowAlphas = new double[windowSize][numberOfWeakLearners];
      windowOutput = new double[windowSize];
      windowTarget = new int[windowSize];
    }
    
    if (errors == null)
    {
      errors = new double[numberOfWeakLearners];
      for (int i = 0; i < numberOfWeakLearners; i++)
      {
        errors[i] = 1.0;
      }
    }
         
    if ((usePretrainedClassifiers) && (!isPrimed))
    {
      throw new RuntimeException("Must prime system if using pretrained classifiers");
    }
    
    isTrained = true;
    
    onlineExamplesCounter++;
    
    double currentLambda = 1;
    int count = 0;
    
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      int k = Poisson.staticNextInt(currentLambda);
      for (int j = 0; j < k; j++)
      {
        weakLearners[i].onlineTrain(trainingData,target);
      }//do k times
      
      int testValue = weakLearners[i].classifySample(trainingData);
      if (testValue == target)  
      {
        weakLearnerCorrect[i] += currentLambda;
        errorValues[i] = ((double) weakLearnerWrong[i] + 1) / (weakLearnerCorrect[i] + weakLearnerWrong[i] + 1);
        currentLambda *= (1.0/(2 - 2*errorValues[i]));
      }//if correct, reduce weight
      else
      {
        weakLearnerWrong[i] += currentLambda;
        errorValues[i] = ((double) weakLearnerWrong[i] + 1) / (weakLearnerCorrect[i] + weakLearnerWrong[i] + 1);
        currentLambda *= (1.0/ (2*errorValues[i]));
      }//else wrong, increase weight
      
      if (examplesOfProtection[i] > 0)
      {
        examplesOfProtection[i]--;
      }
    }//for numberOfWeakLearners
    
    //store example if windowing, averaging, or interval sampling
    
    //windowing is simple, just use an array and back copy like the removal method does
    
    //interval sampling is equally simple, just count and then put the nth example in an arraylist.
    
    //averaging requires both a window array (for the most recent examples) and an arraylist for the previous averaged examples
    //when the array fills up copy it out into the array list in compressed form
    if (isWindowed)
    {
      //window code
      int[] newExample = new int[numberOfWeakLearners];
      double[] newAlpha = new double[numberOfWeakLearners];
      int prediction = 0;
      for (int i = 0; i < numberOfWeakLearners; i++)
      {
        newExample[i] = weakLearners[i].classifySample(trainingData);
        newAlpha[i] = 0.5 * Math.log((1.0-errorValues[i]) / errorValues[i]);
      }
      prediction = ensembleClassification(newExample);
      
      for (int i = windowSize-2; i >= 0; i--)
      {
        //because a double array is an array of arrays this assignment is ok
        windowAlphas[i+1] = windowAlphas[i];
        windowLearners[i+1] = windowLearners[i];
        windowOutput[i+1] = windowOutput[i];
        windowTarget[i+1] = windowTarget[i];
      }
      
      windowAlphas[0] = newAlpha;
      windowLearners[0] = newExample;
      windowOutput[0] = prediction;
      windowTarget[0] = target;
      
      if (windowCounter < windowSize)
      {
        windowCounter++;
      }
    }//if windowing
    else if (isIntervalSampling)
    {
      //interval sampling
    }//else if interval sampling
    else if (isWindowAveraging)
    {
      //window averaging
    }//else if window averaging
    
    if (onlineExamplesCounter == examplesBeforeBackwardStep)
    {
      onlineExamplesCounter = 0;
      int classifierToRemove = -1;
      
      //figure out which classifier to remove, store it in classifierToRemove
      classifierToRemove = selectWeakLearnerForRemoval();
      
      if (classifierToRemove != -1)
      {        
        //remove classifier
        //System.out.println("Removed classifier using feature " + weakLearners[classifierToRemove].getFeatureList()[0] + " at position " + classifierToRemove);
        removeWeakLearner(classifierToRemove);
        addWeakLearner();
      }//if found one to remove
    }//if backwards step time
    
    
    
    /* put other option code here.
    //usePretrainedClassifiers - back & forward;
    //storeOldClassifiers - back & forward;
    done - protectNewClassifiers - back;
    done - examplesToProtect - back;
    //floatNumberOfClassifiers - back & forward;
    //maxNumClassifiers - forward;
    //minNumClassifiers - back;
    //float down can be just a probability of removal and not replacement.
    //float up can be if at a removal step there is no classifier which can be removed to improve the ensemble.
    
    ////decayOldClassifiers - backward step;
    ////examplesToDecay - backward step & forward step;
    done - windowSize - each itr;
    isWindowAveraging - backward step & each itr;
    done - isWindowed - backward step & each itr;
    isIntervalSampling - backward step & each itr;
    */
    
    return classifySample(trainingData);
    //return 0.0;
  }
  
  private int selectWeakLearnerForRemoval()
  {
    int classifierToRemove = -1;
    
    if (isWindowed)
    {
      //window code
                
        //new version, uses current classification error
        int currentError = 0;
        double [] alpha = new double[numberOfWeakLearners];
        for (int i = 0; i < numberOfWeakLearners; i++)
        {
          alpha[i] = 0.5 * Math.log((1.0-errorValues[i]) / errorValues[i]);
        }
        //end new version
        
        //modify this to recalculate the classifications based on the stored alpha maps
        for (int i = 0; i < windowCounter; i++)
        {
          int currentOutput = 0;
          double posClass = 0.0;
          double negClass = 0.0;
          for (int j = 0; j < numberOfWeakLearners; j++)
          {
            if (windowLearners[i][j] == 1)
            {
              posClass += windowAlphas[i][j];
            }
            else
            {
              negClass += windowAlphas[i][j];
            }
          }
          if (posClass > negClass)
          {
            currentOutput = IClassifier.positiveClass;
          }
          else
          {
            currentOutput = IClassifier.negativeClass;
          }
          
          if (currentOutput!=windowTarget[i])
          {
            currentError += 1;
          }
        }
        
        double [] alphaMap = new double[numberOfWeakLearners-1];
        int trainError;
        int l = 0;
        
        int[][] classifications = new int[windowCounter][numberOfWeakLearners-1];
        
        for (int j = 1; j < numberOfWeakLearners; j++)
        {
          alphaMap[l] = 0.5 * Math.log((1.0-errorValues[j]) / errorValues[j]);
          for (int k = 0; k < windowCounter; k++)
          {
            classifications[k][l] = windowLearners[k][j];
          }
          l++;
        }
        
        for (int i = 0; i < numberOfWeakLearners; i++)
        {
          if (i > 0)
          {
            int iMinus = i-1;
            alphaMap[iMinus] = alpha[iMinus];
            for (int k = 0; k < windowCounter; k++)
            {
              classifications[k][iMinus] = windowLearners[k][iMinus];
            }
          }
          trainError = 0;
          
          for (int m = 0; m < windowCounter; m++)
          {
            if (!signCheck(trainingClassification(classifications[m],alphaMap),windowTarget[m]))
            {
              trainError += 1;
            }
          }//work out error value
          
          if (protectNewClassifiers)
          {
            //should move protection into selectWeakLearnerForRemoval()
            if (examplesOfProtection[i] == 0)
            {
              if (trainError < currentError)
              {
                currentError = trainError;
                classifierToRemove = i;
              }
            }//if unprotected
          }//if protecting
          else
          {
            if (trainError < currentError)
            {
              currentError = trainError;
              classifierToRemove = i;
            }
          }//else not protecting
        }//iterate over the weakLearnersSet
    }//if windowing
    else if (isIntervalSampling)
    {
      //interval sampling
    }//else if interval sampling
    else if (isWindowAveraging)
    {
      //window averaging
    }//else if window averaging
    
    return classifierToRemove;
  }//selectWeakLearnerForRemoval()
  
  private void removeWeakLearner(int classifierToRemove)
  {
    numberOfWeakLearners--;
    for (int i = classifierToRemove; i < numberOfWeakLearners; i++)
    {
      int iPlus = i+1;
      weakLearners[i] = weakLearners[iPlus];
      weakLearnerCorrect[i] = weakLearnerCorrect[iPlus];
      weakLearnerWrong[i] = weakLearnerWrong[iPlus];
      errorValues[i] = errorValues[iPlus];
      
      for (int j = 0; j < windowCounter; j++)
      {
        windowLearners[j][i] = windowLearners[j][iPlus];
        windowAlphas[j][i] = windowAlphas[j][iPlus];
      }
    }
    
    for (int j = 0; j < windowCounter; j++)
    {
      windowLearners[j][numberOfWeakLearners] = 0;
      windowAlphas[j][numberOfWeakLearners] = 0.0;
    }
  }//removeWeakLearner(int)
  
  private void addWeakLearner()
  {
    //when adding a new classifier
    examplesOfProtection[numberOfWeakLearners] = examplesToProtect;
    
    if (usePretrainedClassifiers)
    {
      //pick a pretrained one and store it in weakLearners[numberOfWeaklearners];
    }//if using pretrainedClassifiers
    else
    {
      weakLearners[numberOfWeakLearners] = rootWeakLearner.copyClassifier();
      weakLearnerCorrect[numberOfWeakLearners] = 0;
      weakLearnerWrong[numberOfWeakLearners] = 0;
      errorValues[numberOfWeakLearners] = 1;
    }//else create new weak learner
    
    numberOfWeakLearners++;
  }//addWeakLearner()
    
  
  private double offlineTrain(IDataset trainingData, int iterations)
  {
    //returns the training error
    double returnVal = 0.0;
    
    numberOfWeakLearners = iterations;
  
    double[] alphaValues = new double[iterations];
    errorValues = new double[iterations];
    weakLearners = new IOnlineClassifier[iterations];
    weakLearnerCorrect = new double[iterations];
    weakLearnerWrong = new double[iterations];
    
    int numberOfSamples = trainingData.getNumberOfSamples();
    
    int [] targets = trainingData.getTargetVector();
    
    double [] oldWeights = trainingData.getDataWeights();
    
    double [] newWeights = trainingData.getDataWeights();
    
    errors = new double[iterations+1];
    
    for (int i = 0; i <= iterations; i++)
    {
      errors[i] = 1.0;
    }
    
    HashMap<IOnlineClassifier,int[]> weakClassifications = new HashMap<IOnlineClassifier,int[]>(iterations);
        
    double posClass = 0;
    double negClass = 0;
    
    for (int i = 0; i < numberOfSamples; i++)
    {
      if (targets[i] == IClassifier.positiveClass)
      {
        posClass = posClass + oldWeights[i];
      }//if positive
      else
      {
        negClass = negClass + oldWeights[i];
      }//else negative
    }
    
    double weightCounter = 0;
    
    for (int i = 0; i < numberOfSamples; i++)
    {
      if (targets[i] == IClassifier.positiveClass)
      {
        newWeights[i] = newWeights[i] / (2*posClass);
      }
      else
      {
        newWeights[i] = newWeights[i] / (2*negClass);
      }
      //weightCounter += newWeights[i];
    }
    
    //System.out.println("Weight counter = " + weightCounter);
     
    //resets the training weights, means an adaboost of adaboosts of weak learners won't work.
    trainingData.setDataWeights(newWeights);
      
    double trainError; 
    double weightNormalise;
    
    ArrayList<IOnlineClassifier> weakLearnersSet = new ArrayList<IOnlineClassifier>(iterations);
    HashMap<IOnlineClassifier,Double> weakLearnerErrorMap = new HashMap<IOnlineClassifier,Double>(iterations);
    HashMap<IOnlineClassifier,Double> weakLearnerCorrectMap = new HashMap<IOnlineClassifier,Double>(iterations);
    HashMap<IOnlineClassifier,Double> weakLearnerWrongMap = new HashMap<IOnlineClassifier,Double>(iterations);
    
    int j = 0;
    int loopCounter = 0;
    
    while ((j < iterations) && (loopCounter < iterations * 10))
    {
      loopCounter++;
    
      //train new classifier
      IOnlineClassifier newWeakLearner = rootWeakLearner.copyClassifier();
      double weakError = newWeakLearner.train(trainingData,weaklearnerIterations);
      double tempAlpha = 0;
      double curWeakLearnerCorrect = 0;
      double curWeakLearnerWrong = 0;
            
      if (weakError < 0.0000000001)
      {
        tempAlpha = 1.0;
        weakError = 0.0000000001;
      }
      else
      {
        tempAlpha = 0.5 * Math.log((1.0-weakError) / weakError);
      }
      
      int[] newWeakClassification = newWeakLearner.classifyDataset(trainingData);
      weakLearnersSet.add(newWeakLearner);
      weakLearnerErrorMap.put(newWeakLearner,new Double(tempAlpha));
      weakClassifications.put(newWeakLearner,newWeakClassification);
      j++;
      
      //System.out.println("j = " + j + ", feature = " + newWeakLearner.getFeatureList()[0] + ", trerr = " + weakError);
      
      //work out system error
      int[][] transposedClassifications = new int[numberOfSamples][j];
     
      IOnlineClassifier[] learnerArray = weakLearnersSet.toArray(new IOnlineClassifier[0]);
      double[] alphaMap = new double[j];
      transposedClassifications = new int[numberOfSamples][j];
      int i = 0;
      for (int m = 0; m < learnerArray.length; m++)
      { 
        int[] array = weakClassifications.get(learnerArray[m]);
        alphaMap[m] = weakLearnerErrorMap.get(learnerArray[m]);
        for (int k = 0; k < array.length; k++)
        {
          transposedClassifications[k][i] = array[k];
        }
        i++;
      }
      
      //now reweight data
      weightNormalise = 0.0;
      trainError = 0.0;
      for (int k = 0; k < numberOfSamples; k++)
      {
        if (signCheck(newWeakClassification[k],targets[k]))
        {
          curWeakLearnerCorrect++;
          //curWeakLearnerCorrect += newWeights[k];
        }
        else
        {
          curWeakLearnerWrong++;
          //curWeakLearnerWrong += newWeights[k];
        }
        double curPrediction = trainingClassification(transposedClassifications[k],alphaMap);
        if (signCheck(curPrediction,targets[k]))
        {
          newWeights[k] = Math.exp(-Math.abs(curPrediction));
        }
        else
        {
          trainError += newWeights[k];
          newWeights[k] = Math.exp(Math.abs(curPrediction));
        }
        weightNormalise += newWeights[k];
      }//work out error value
      
      for (int k = 0; k < numberOfSamples; k++)
      {
        newWeights[k] = newWeights[k] / weightNormalise;
      }//normalise weights
      
      weakLearnerCorrectMap.put(newWeakLearner,new Double(curWeakLearnerCorrect));
      weakLearnerWrongMap.put(newWeakLearner,new Double(curWeakLearnerWrong));
      
      /*
      int[] newWeakClassification = newWeakLearner.classifyDataset(trainingData);
      weakLearnersSet.add(newWeakLearner);
      weakLearnerErrorMap.put(newWeakLearner,new Double(weakError));
      weakClassifications.put(newWeakLearner,newWeakClassification);
      j++;
      */
      
      //check error value, and store
      if (trainError < errors[j])
      {
        errors[j] = trainError;
      }
      
      //System.out.println("Train error is " + trainError);
      
      boolean removedLearner = false;
      //now do conditional exclusion
      while(j > 1)
      {
        loopCounter++;
        //work out error things
        double currentError = 1.0;
        learnerArray = weakLearnersSet.toArray(new IOnlineClassifier[0]);
        IClassifier currentWeak = rootWeakLearner.copyClassifier();
        alphaMap = new double[j];
        
        for (int k = 0; k < learnerArray.length; k++)
        {
          alphaMap[k] = weakLearnerErrorMap.get(learnerArray[k]);
        }
        
        int sameErrorCounter = 0;
        transposedClassifications = new int[numberOfSamples][j-1];
        for (int k = 0; k < learnerArray.length; k++)
        {
          i = 0;
          for (int m = 0; m < learnerArray.length; m++)
          {
            if (m != k)
            {
              int[] array = weakClassifications.get(learnerArray[m]);
              for (int n = 0; n < array.length; n++)
              {
                transposedClassifications[n][i] = array[n];
              }
              i++;
            }
          }
          
          trainError = 0.0;
          for (int m = 0; m < numberOfSamples; m++)
          {
            if (!signCheck(trainingClassification(transposedClassifications[m],alphaMap),targets[m]))
            {
              trainError += newWeights[m];
            }
          }//work out error value
          
          if (trainError < currentError)
          {
            currentError = trainError;
            currentWeak = learnerArray[k];
            //System.out.println("train error at removal is " + trainError);
          }
          else if (trainError == currentError)
          {
            sameErrorCounter++;
          }
        }//iterate over the weakLearnersSet
        
        //if (sameErrorCounter != 0)
        //  System.out.println("SameErrorCounter = " + sameErrorCounter);
        
        if (currentError < errors[j-1])
        {
          weakLearnersSet.remove(currentWeak);
          weakLearnerErrorMap.remove(currentWeak);
          weakClassifications.remove(currentWeak);
          weakLearnerCorrectMap.remove(currentWeak);
          weakLearnerWrongMap.remove(currentWeak);
          j--;
          
          //System.out.println("Removed feature " + currentWeak.getFeatureList()[0] + ", with trerr = " + currentError + ", curMinErr = " + errors[j]);
          
          errors[j] = currentError;
          removedLearner = true;
          //System.out.println("Removed learner, counter = " + loopCounter);
        }//if currentError less than min error
        else
        {
          if (removedLearner)
          {
            learnerArray = weakLearnersSet.toArray(new IOnlineClassifier[0]);
            alphaMap = new double[j];
            transposedClassifications = new int[numberOfSamples][j];
            i = 0;
            for (int m = 0; m < learnerArray.length; m++)
            { 
              int[] array = weakClassifications.get(learnerArray[m]);
              alphaMap[m] = weakLearnerErrorMap.get(learnerArray[m]);
              for (int k = 0; k < array.length; k++)
              {
                transposedClassifications[k][i] = array[k];
              }
              i++;
            }
            
            //now reweight data
            weightNormalise = 0.0;
            for (int k = 0; k < numberOfSamples; k++)
            {
              double curPrediction = trainingClassification(transposedClassifications[k],alphaMap);
              if (signCheck(curPrediction,targets[k]))
              {
                newWeights[k] = Math.exp(-Math.abs(curPrediction));
              }
              else
              {
                newWeights[k] = Math.exp(Math.abs(curPrediction));
              }
              weightNormalise += newWeights[k];
            }//work out error value
      
            for (int k = 0; k < numberOfSamples; k++)
            {
              newWeights[k] = newWeights[k] / weightNormalise;
            }//normalise weights
          }//if removed a weak learner
          break;
        }//else currentError not less than min error
      }//while conditional exclusion
      //System.out.println("iteration = " + loopCounter + ", j = " + j);
    }//while j < iterations
    
    isPrimed = true;
    isTrained = true;
    
    //otherwise it uses incorrect values for the training weights
    trainingData.setDataWeights(oldWeights);
    
    System.out.println("Number of iterations = " + loopCounter);
    
    weakLearners = weakLearnersSet.toArray(new IOnlineClassifier[0]);
    alphaValues = new double[iterations];
    
    for (int i = 0; i < iterations; i++)
    {
      alphaValues[i] = weakLearnerErrorMap.get(weakLearners[i]);
      weakLearnerCorrect[i] = weakLearnerCorrectMap.get(weakLearners[i]);
      weakLearnerWrong[i] = weakLearnerWrongMap.get(weakLearners[i]);
      errorValues[i] = (double) weakLearnerWrong[i] / (weakLearnerWrong[i] + weakLearnerCorrect[i]);
    }//copy out alpha values
    
    returnVal = test(trainingData);
    
    return returnVal;
  }
  
  /**
  * Is the main offline test function, which takes an dataset and returns the testing error
  * on that dataset
  * 
  * @param  testingData    A testing dataset
  */
  public double test(IDataset testingData)
  {
    double returnVal = 0.0;
    if (isTrained)
    {
      //System.out.println("Number of weak learners = " + numberOfWeakLearners);
      //run test
      int numberOfTestSamples = testingData.getNumberOfSamples();
      int[] targetVector = testingData.getTargetVector();
      double[] testWeights = testingData.getDataWeights();
      double totalWeight = 0.0;
      int count = 0;
      
      for (int i = 0; i < numberOfTestSamples; i++)
      {
        double[] currentSample = testingData.getSampleDouble(i,false);
        int finalClassification = classifySample(currentSample);
        if (finalClassification != targetVector[i])
        {
          returnVal = returnVal + testWeights[i];
          count = count + 1;
        }
        totalWeight = totalWeight + testWeights[i];
      }
            
      //System.out.println("Total weight = " + totalWeight);
      //System.out.println("Actual Percentage = " + (double) count / numberOfTestSamples);
      //returnVal = returnVal / numberOfTestSamples;
    }
    else
    {
      returnVal = -1.0;
    }//else untrained, return -1
    return returnVal;
  }//test(IDataset)
  
  
  /**
  * Returns the classifications made by the trained classifier on the testing data
  * 
  * @param  testingData    A testing dataset
  */
  public int[] classifyDataset(IDataset testingData)
  {
    int [] classification;
    if (isTrained)
    {
      //run test
      int numberOfTestSamples = testingData.getNumberOfSamples();
      
      classification = new int[numberOfTestSamples];
      
      int [] learnerClassification = new int[numberOfWeakLearners];
      for (int i = 0; i < numberOfTestSamples; i++)
      {
        double[] currentSample = testingData.getSampleDouble(i,false);
        classification[i] = classifySample(currentSample);
      }
    }
    else
    {
      classification = null;
    }//else untrained, return null
    return classification;
  }//classify(IDataset)
  
  /**
  * Returns the classification for a single example
  * 
  * @param  sample        A testing sample
  */  
  public int classifySample(double[] sample)
  {
    int [] learnerClassification = new int[numberOfWeakLearners];
    for (int j = 0; j < numberOfWeakLearners; j++)
    {
      learnerClassification[j] = weakLearners[j].classifySample(sample);
    }
    return ensembleClassification(learnerClassification);
  }
    
  private int ensembleClassification(int[] learnerClassifications)
  {
    double posCounter = 0.0;
    double negCounter = 0.0;
    int finalClass;
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      if (learnerClassifications[i] == 1)
      {
        posCounter = posCounter + Math.log((1-errorValues[i])/ errorValues[i]);
      }
      else
      {
        negCounter = negCounter + Math.log((1-errorValues[i])/ errorValues[i]);
      }
    }
    
    if (posCounter > negCounter)
    {
      finalClass = IClassifier.positiveClass;
    }
    else
    {
      finalClass = IClassifier.negativeClass;
    }
    
    return finalClass;
  }//ensembleClassification(int [])
  
  private double trainingClassification(int[] learnerClassifications, double[] alphaMapping)
  {
    double counter = 0.0;
    int finalClass;
    for (int i = 0; i < learnerClassifications.length; i++)
    {
      counter = counter + (learnerClassifications[i] * alphaMapping[i]);
    }
    
    return counter;
  }//ensembleClassification(int [])
  
  private boolean signCheck(double testValue, int answer)
  {
    if ((testValue > 0) && (answer > 0))
    {
      return true;
    }
    else if ((testValue < 0) && (answer < 0))
    {
      return true;
    }
    else
    {
      return false;
    }
  }//signCheck(double,int)


  /**
  * Returns the list of features used by each weak learner (useful for decision stumps)
  * 
  */  
  public int[] getFeatureList()
  {
    if (isTrained)
    {
      int listLength = 0;
      int totalLength = 0;
      int [] tmp = weakLearners[0].getFeatureList();
      int [] returnArray = new int[0];
      
      if (tmp.length > 0)
      {
        listLength = tmp.length;
        totalLength = tmp.length * numberOfWeakLearners;
        returnArray = new int[totalLength];
        
        for (int i = 0; i < numberOfWeakLearners; i++)
        {
          tmp = weakLearners[i].getFeatureList();
          for (int j = 0; j < tmp.length; j++)
          {
            returnArray[i*listLength + j] = tmp[j];
          }
        }
      }
      
      return returnArray;
    }//if trained
    else
    {
      return new int[0];
    }//else return empty list
  }
  
  public boolean getIsTrained()
  {
    return isTrained;
  }//getIsTrained
  
  public boolean supportsWeightedData()
  {
    return true;
  }//supportsWeightedData
  
  public boolean supportsMultiClassData()
  {
    return false;//rootWeakLearner.supportsMultiClassData();
  }
}//class AdaBoost
