package JavaML;

import cern.jet.random.Poisson;
import java.util.Random;


/**
 * An implementation of Online Boosting.
 * When using offline datasets, performs exactly like AdaBoost
 * Can be primed by training on a small subset of data like AdaBoost
 * and then commencing online learning updates.
 * Executing the train() function resets the classifier before training
 *  
 * @author apocock
 */
public class OnlineAdaBoost implements IOnlineClassifier
{
  private int numberOfWeakLearners;
  
  private double[] errorValues;
  private IOnlineClassifier[] weakLearners;
  private double[] weakLearnerCorrect;
  private double[] weakLearnerWrong;
  
  private IOnlineClassifier rootWeakLearner;
  private String weakLearnerArgs;
  private int weakLearnerIterations;
  
  private boolean isPrimed;
  private boolean isTrained;
      
  private Random rndNumberGen;
  private long rndSeed;
  
  
  /**
  * Constructor, needs to be supplied with a weak learner before training
  *
  */
  public OnlineAdaBoost()
  {
    numberOfWeakLearners = 0;
    
    errorValues = new double[0];
    weakLearners = new IOnlineClassifier[0];
    weakLearnerCorrect = new double[0];
    weakLearnerWrong = new double[0];
    
    rootWeakLearner = null;
    weakLearnerArgs = "";
    
    isPrimed = false;
    isTrained = false;
    
    rndSeed = System.currentTimeMillis();
    rndNumberGen = new Random(rndSeed);
  }//OnlineAdaBoost constructor
  
  /**
  * Constructor, initialises everything required
  * @param  numLearners   Is the number of weak learners to use
  * @param  rootLearner   Is an IOnlineClassifier which all the weak learners are constructed from
  * @param  learnerArgs   Are any arguments that weak learner needs to be constructed with
  * @param  priming       Specifies if the classifier is going to be trained offline first
  */
  public OnlineAdaBoost(int numLearners, IOnlineClassifier rootLearner, String learnerArgs, boolean priming)
  {
    numberOfWeakLearners = numLearners;
    
    errorValues = new double[numberOfWeakLearners];
    weakLearners = new IOnlineClassifier[numberOfWeakLearners];
    weakLearnerCorrect = new double[numberOfWeakLearners];
    weakLearnerWrong = new double[numberOfWeakLearners];
    
    rootWeakLearner = rootLearner;
    weakLearnerArgs = learnerArgs;
    weakLearnerIterations = 1;
    
    isPrimed = priming;
    isTrained = false;
    
    rndSeed = System.currentTimeMillis();
    rndNumberGen = new Random(rndSeed);
    
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      weakLearners[i] = rootWeakLearner.copyClassifier();
    }
  }//constructor
  
  /**
  * Constructor, initialises everything required
  * @param  numLearners   Is the number of weak learners to use
  * @param  rootLearner   Is an IOnlineClassifier which all the weak learners are constructed from
  * @param  learnerArgs   Are any arguments that weak learner needs to be constructed with
  * @param  learnerIterations If the weak learner needs the iterations parameter (e.g. for boosting a neural net)
  * @param  priming       Specifies if the classifier is going to be trained offline first
  */
  public OnlineAdaBoost(int numLearners, IOnlineClassifier rootLearner, int learnerIterations, String learnerArgs, boolean priming)
  {
    numberOfWeakLearners = numLearners;
    
    errorValues = new double[numberOfWeakLearners];
    weakLearners = new IOnlineClassifier[numberOfWeakLearners];
    weakLearnerCorrect = new double[numberOfWeakLearners];
    weakLearnerWrong = new double[numberOfWeakLearners];
    
    
    rootWeakLearner = rootLearner;
    weakLearnerArgs = learnerArgs;
    weakLearnerIterations = learnerIterations;
    
    isPrimed = priming;
    isTrained = false;
    
    rndSeed = System.currentTimeMillis();
    rndNumberGen = new Random(rndSeed);
    
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      weakLearners[i] = rootWeakLearner.copyClassifier();
    }
  }//constructor
  
  /**
  * Copy constructor
  */
  public IOnlineClassifier copyClassifier()
  {
    IOnlineClassifier newAdaBoost = new OnlineAdaBoost(numberOfWeakLearners, rootWeakLearner, weakLearnerIterations, weakLearnerArgs, isPrimed);
    return newAdaBoost;
  }//copyClassifier()  
  
  /**
  * setupClassifier is used to modify classifiers after they have been constructed through the copy constructor,
  * at that point you could either do reflection to figure out the class and reconstruct it, or you could pass it
  * a string of arguments and have it reconfigure itself.
  *
  * This classifier accepts "WeakLearner=<CLASSNAME> NumLearners=<int> WeakLearnerIterations=<int>"
  *
  * @param  arguments   a space separated list of parameter=value tuples.
  */
  public void setupClassifier(String arguments)
  {
    String [] argsList = arguments.split(" ");
    try
    {
      for (int i = 0; i < argsList.length; i++)
      {
        String [] arg = argsList[i].split("=");
        if (arg.length == 2)
        {
          if (arg[0].equals("WeakLearner"))
          {
            Class learnerClass = null;
            
            System.out.println(arg[1]);
            
            learnerClass = Class.forName(arg[1]);
            
            Object learnerObject = learnerClass.newInstance();
            
            rootWeakLearner = (IOnlineClassifier) learnerClass.cast(learnerObject);
          }//if weaklearner argument
          else if (arg[0].equals("NumLearners"))
          {
            int numLearners = Integer.parseInt(arg[1]);
            numberOfWeakLearners = numLearners;
            errorValues = new double[numberOfWeakLearners];
            weakLearners = new IOnlineClassifier[numberOfWeakLearners];
            weakLearnerCorrect = new double[numberOfWeakLearners];
            weakLearnerWrong = new double[numberOfWeakLearners];
          }
          else if (arg[0].equals("WeakLearnerIterations"))
          {
            weakLearnerIterations = Integer.parseInt(arg[1]);
          }
        }//if arg formats
      }//for number of arguments
    }//try
    catch (NumberFormatException e)
    {
      System.out.println("Incorrectly formatted number in arguments");
    }
    catch (ArrayIndexOutOfBoundsException e)
    {
      System.out.println("Incorrectly formatted argument, missing equals");
    }
    catch (ClassCastException e)
    {
      System.out.println("Class casting exception, the WeakLearner did not implement IOnlineClassifier");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
    catch (ClassNotFoundException e)
    {
      System.out.println("WeakLearner class not found");
    }
    catch (InstantiationException e)
    {
      System.out.println("Instantiation exception on creation of new WeakLearner");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
    catch (Exception e)
    {
      System.out.println("Other generic exception in setupClassifier");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
  }//setupClassifier


  /**
  * Is the main offline train function, which takes an dataset and trains on it using offline AdaBoost
  * 
  *
  * @param  trainingData    A training dataset
  * @param  iterations      overrides the inbuilt number of weak leaners, and creates <code>iterations</code> weak learners
  */
  public double train(IDataset trainingData, int iterations)
  {
    double returnVal = 0.0;
    if ((rootWeakLearner == null) || (iterations < 1))
    {
      returnVal = -1.0;
    }
    else
    {
      returnVal = offlineTrain(trainingData,iterations);
    }
    return returnVal;
  }
  
  /**
  * Is the online train function, which takes a single example and trains on it
  * then returns the classification.
  *
  * @param  trainingData    A training sample
  * @param  target          class label for the sample
  */
  public int onlineTrain(double[] trainingData, int target)
  {
    if (numberOfWeakLearners == 0)
    {
      errorValues = new double[numberOfWeakLearners];
      weakLearners = new IOnlineClassifier[numberOfWeakLearners];
      weakLearnerCorrect = new double[numberOfWeakLearners];
      weakLearnerWrong = new double[numberOfWeakLearners];
    }
    
    isTrained = true;
    
    double currentLambda = 1;
    int count = 0;
    
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      int k = Poisson.staticNextInt(currentLambda);
      //System.out.println("i = " + i + ", k = " + k + ", lambda = " + currentLambda);
      for (int j = 0; j < k; j++)
      {
        weakLearners[i].onlineTrain(trainingData,target);
      }//do k times
      
      int testValue = weakLearners[i].classifySample(trainingData);
      if (testValue == target)
      {
        weakLearnerCorrect[i] += currentLambda;
        //errorValues[i] = (double) weakLearnerWrong[i] / (weakLearnerCorrect[i] + weakLearnerWrong[i]);
        errorValues[i] = ((double) weakLearnerWrong[i] + 1) / (weakLearnerCorrect[i] + weakLearnerWrong[i] + 1);
        //System.out.println("true errorValues["+i+"] = " + errorValues[i]);
        currentLambda *= (1.0/(2 - 2*errorValues[i]));
      }//if correct, reduce weight
      else
      {
        weakLearnerWrong[i] += currentLambda;
        //errorValues[i] = (double) weakLearnerWrong[i] / (weakLearnerCorrect[i] + weakLearnerWrong[i]);
        errorValues[i] = ((double) weakLearnerWrong[i] + 1) / (weakLearnerCorrect[i] + weakLearnerWrong[i] + 1);
        //System.out.println("false errorValues["+i+"] = " + errorValues[i]);
        currentLambda *= (1.0/ (2*errorValues[i]));
      }//else wrong, increase weight
    }//for numberOfWeakLearners
    
    return classifySample(trainingData);
  }
  
  private double offlineTrain(IDataset trainingData, int iterations)
  {
    //returns the training error
    double returnVal = 0.0;
    
    numberOfWeakLearners = iterations;
  
    errorValues = new double[iterations];
    weakLearners = new IOnlineClassifier[iterations];
    weakLearnerCorrect = new double[iterations];
    weakLearnerWrong = new double[iterations];
    
    double [] alphaValues = new double[iterations];
    
    int numberOfSamples = trainingData.getNumberOfSamples();
    
    int [] targets = trainingData.getTargetVector();
    
    double [] oldWeights = trainingData.getDataWeights();
    
    double [] newWeights = trainingData.getDataWeights();
    
    trainingData.setDataWeights(newWeights);
      
    double [] sample;
    int classificationResult;
    double trainError;
    double distributionCount;
    
    for (int i = 0; i < iterations; i++)
    {
      weakLearners[i] = rootWeakLearner.copyClassifier();
      
      trainError = weakLearners[i].train(trainingData,weakLearnerIterations);
      
      if (trainError == 0.0)
      {
        alphaValues[i] = 1.0;
      }
      else
      {
        alphaValues[i] = 0.5 * Math.log((1.0-trainError) / trainError);
      }
      
      distributionCount = 0.0;
      
      double expValNeg = Math.exp(-alphaValues[i]);
      double expValPos = Math.exp(alphaValues[i]);
      
      for (int j = 0; j < numberOfSamples; j++)
      {
        sample = trainingData.getSampleDouble(j,false);
        classificationResult = weakLearners[i].classifySample(sample);
        
        //slight hack here, results in the same answer for binary class systems
        //will modify behaviour in multiple class systems in odd ways
        double expVal;
        
        if (classificationResult == targets[j])
        {
          expVal = expValNeg;
          weakLearnerCorrect[i] += newWeights[j];
        }
        else
        {
          expVal = expValPos;
          weakLearnerWrong[i] += newWeights[j];
        }
        
        newWeights[j] = newWeights[j] * expVal;
        distributionCount = distributionCount + newWeights[j];
      }//recalculate sample weights
      
      for (int j = 0; j < numberOfSamples; j++)
      {
        newWeights[j] = newWeights[j] / distributionCount;
      }//normalising the weights
      
      errorValues[i] = (double) weakLearnerWrong[i] / (weakLearnerCorrect[i] + weakLearnerWrong[i]);
    }//generate the weak learners
        
    isTrained = true;
    isPrimed = true;
    
    //otherwise it uses incorrect values for the training weights
    trainingData.setDataWeights(oldWeights);
        
    returnVal = test(trainingData);
    
    return returnVal;
  }
  

  /**
  * Is the main offline test function, which takes an dataset and returns the testing error
  * on that dataset
  * 
  * @param  testingData    A testing dataset
  */
  public double test(IDataset testingData)
  {
    double returnVal = 0.0;
    if (isTrained)
    {
      //System.out.println("Number of weak learners = " + numberOfWeakLearners);
      //run test
      int numberOfTestSamples = testingData.getNumberOfSamples();
      int[] targetVector = testingData.getTargetVector();
      double[] testWeights = testingData.getDataWeights();
      double totalWeight = 0.0;
      int count = 0;
      
      for (int i = 0; i < numberOfTestSamples; i++)
      {
        double[] currentSample = testingData.getSampleDouble(i,false);
        int finalClassification = classifySample(currentSample);
        if (finalClassification != targetVector[i])
        {
          returnVal = returnVal + testWeights[i];//1.0;
          count = count + 1;
        }
        totalWeight = totalWeight + testWeights[i];
      }
      
      //System.out.println("###### Temp = " + (returnVal / totalWeight));
      
      //System.out.println("Total weight = " + totalWeight);
      //System.out.println("Actual Percentage = " + (double) count / numberOfTestSamples);
      //returnVal = returnVal / numberOfTestSamples;
    }
    else
    {
      returnVal = -1.0;
    }//else untrained, return -1
    return returnVal;
  }//test(IDataset)
  
  /**
  * Returns the classifications made by the trained classifier on the testing data
  * 
  * @param  testingData    A testing dataset
  */
  public int[] classifyDataset(IDataset testingData)
  {
    int [] classification;
    if (isTrained)
    {
      //run test
      int numberOfTestSamples = testingData.getNumberOfSamples();
      
      classification = new int[numberOfTestSamples];
      
      int [] learnerClassification = new int[numberOfWeakLearners];
      for (int i = 0; i < numberOfTestSamples; i++)
      {
        double[] currentSample = testingData.getSampleDouble(i,false);
        classification[i] = classifySample(currentSample);
      }
    }
    else
    {
      classification = null;
    }//else untrained, return null
    return classification;
  }//classify(IDataset)
  
  /**
  * Returns the classification for a single example
  * 
  * @param  sample        A testing sample
  */  
  public int classifySample(double[] sample)
  {
    int [] learnerClassification = new int[numberOfWeakLearners];
    for (int j = 0; j < numberOfWeakLearners; j++)
    {
      learnerClassification[j] = weakLearners[j].classifySample(sample);
    }
    return ensembleClassification(learnerClassification);
  }
  
  private int ensembleClassification(int[] learnerClassifications)
  {
    double posCounter = 0.0;
    double negCounter = 0.0;
    int finalClass;
    for (int i = 0; i < numberOfWeakLearners; i++)
    {
      if (learnerClassifications[i] == 1)
      {
        posCounter = posCounter + Math.log((1-errorValues[i])/ errorValues[i]);
      }
      else
      {
        negCounter = negCounter + Math.log((1-errorValues[i])/ errorValues[i]);
      }
    }
    
    if (posCounter > negCounter)
    {
      finalClass = IClassifier.positiveClass;
    }
    else
    {
      finalClass = IClassifier.negativeClass;
    }
    
    return finalClass;
  }//ensembleClassification(int [])
   
  public void setRandomSeed(long newSeed)
  {
    rndNumberGen.setSeed(newSeed);
    rndSeed = newSeed;
  }  

  /**
  * Returns the list of features used by each weak learner (useful for decision stumps)
  * 
  */  
  public int[] getFeatureList()
  {
    if (isTrained)
    {
      int listLength = 0;
      int totalLength = 0;
      int [] tmp = weakLearners[0].getFeatureList();
      int [] returnArray = new int[0];
           
      if (tmp.length > 0)
      {
        listLength = tmp.length;
        totalLength = tmp.length * numberOfWeakLearners;
        returnArray = new int[totalLength];
        
        for (int i = 0; i < numberOfWeakLearners; i++)
        {
          tmp = weakLearners[i].getFeatureList();
          for (int j = 0; j < tmp.length; j++)
          {
            returnArray[i*listLength + j] = tmp[j];
          }
        }
      }
      
      return returnArray;
    }//if trained
    else
    {
      return new int[0];
    }//else return empty list
  }
  
  public boolean getIsTrained()
  {
    return isTrained;
  }//getIsTrained
  
  public boolean supportsWeightedData()
  {
    return false;
  }//supportsWeightedData
  
  public boolean supportsMultiClassData()
  {
    return false;
  }
}//class AdaBoost
